import arzumify.cursedenergy.api.EnergyProvider;
import arzumify.cursedenergy.api.EnergyReceiver;
import net.fabricmc.fabric.api.event.lifecycle.v1.ServerBlockEntityEvents;
import net.fabricmc.fabric.api.event.lifecycle.v1.ServerTickEvents;
import net.minecraft.block.entity.BlockEntity;
import net.minecraft.server.MinecraftServer;
import net.minecraft.server.world.ServerWorld;
import net.minecraft.util.math.Vec3i;

import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import static arzumify.cursedenergy.api.SearchForNeighbours.searchForNeighbours;

/**
 * SimpleCable, but implemented using only the API and not the actual block entity for testing purposes.
 * All fields are public for testing purposes.
 */
public class SimpleCodeOnlyCable implements EnergyProvider, EnergyReceiver, ServerBlockEntityEvents.Unload, ServerBlockEntityEvents.Load, ServerTickEvents.EndTick {
    public final long capacity;
    public final long inputRate;
    public final long outputRate;
    public final String name;
    public final Set<EnergyProvider> providers = ConcurrentHashMap.newKeySet();
    public final Set<EnergyProvider> blackListed = ConcurrentHashMap.newKeySet();
    public final Set<EnergyReceiver> receivers = ConcurrentHashMap.newKeySet();
    public long energy = 0;

    public SimpleCodeOnlyCable(Vec3i pos, long capacity, long inputRate, long outputRate, String name) {
        this.capacity = capacity;
        this.inputRate = inputRate;
        this.outputRate = outputRate;
        this.name = name;
    }

    public static void tick(SimpleCodeOnlyCable Cable) {
        System.out.println(Cable.name + " ticking");
        // Split the energy between all receivers
        var energyToGive = Cable.energy / Cable.receivers.size();

        for (EnergyReceiver receiver : Cable.receivers) {
            var taken = receiver.receive(energyToGive, Cable);
            Cable.energy -= taken;
            if (receiver instanceof EnergyProvider) {
                Cable.blackListed.add((EnergyProvider) receiver);
            }
        }
    }

    @Override
    public long receive(long amount, EnergyProvider provider) {
        if (!blackListed.contains(provider)) {
            System.out.println(name + " receiving " + amount + " energy");
            long received = Math.min(Math.min(inputRate, amount), capacity - energy);
            energy += received;
            System.out.println(name + " received " + received + " energy");
            blackListed.add(provider);
            return received;
        } else {
            return 0;
        }
    }

    @Override
    public void ready(EnergyReceiver receiver) {
        System.out.println(name + " found other receiver");
        receivers.add(receiver);
        receiver.ready(this);
    }

    @Override
    public void ready(EnergyProvider provider) {
        System.out.println(name + " found other provider");
        providers.add(provider);
    }

    @Override
    public void unready(EnergyReceiver receiver) {
        System.out.println(name + " lost other receiver");
        receivers.remove(receiver);
    }

    @Override
    public void unready(EnergyProvider provider) {
        System.out.println(name + " lost other provider");
        providers.remove(provider);
    }

    @Override
    public void onLoad(BlockEntity blockEntity, ServerWorld serverWorld) {
        var energyBlocks = searchForNeighbours(serverWorld, blockEntity.getPos());
        for (var block : energyBlocks) {
            if (block instanceof EnergyProvider) {
                ((EnergyProvider) block).ready(this);
            }
            if (block instanceof EnergyReceiver) {
                ready((EnergyReceiver) block);
            }
        }
    }

    @Override
    public void onUnload(BlockEntity blockEntity, ServerWorld serverWorld) {
        for (var provider : providers) {
            provider.unready(this);
        }
        providers.clear();
        for (var receiver : receivers) {
            receiver.unready(this);
        }
        receivers.clear();
    }

    @Override
    public void onEndTick(MinecraftServer server) {
        blackListed.clear();
    }
}
